class Document:
  VALID_EXTENSIONS = ('.ai', '.bmp', '.c', '.cpp', '.csv', '.doc', '.docx', 
                      '.eps', '.gif', '.h', '.ico', '.jp2', '.jpeg', '.jpg', 
                      '.m', '.odf', '.pdf', '.php', '.png', '.pps', '.ppsx', 
                      '.ppt', '.pptx', '.ps', '.psd', '.py', '.rtf', '.tif', 
                      '.tiff', '.txt', '.xls', '.xlsx', '.xml', '.xps')

  PORTRAIT = 'portrait'
  LANDSCAPE = 'landscape'

  TWO_SIDED_LONG = 'two-sided-long-edge'
  TWO_SIDED_SHORT = 'two-sided-short-edge'
  ONE_SIDED = 'one-sided'

  LETTER = 'letter'
  TABLOID = 'tabloid'
  CUSTOM = 'custom'

  ON = 'on'
  OFF = 'off'

  filePath = None
  params = {'copies': 1, 
            'scale': ON, 
            'sides': TWO_SIDED_LONG, 
            'range': None, 
            'orientation': PORTRAIT,
            'size': LETTER,
            'size_length': None,
            'size_width': None
            }

  def __init__(self, filePathIn):
    self.filePath = filePathIn
      
  def __str__(self):
    return self.filePath



import sys
import os
import json
from .settings import webloginBaseURL, userBaseURL, guestBaseURL  # Base URLs
import requests # Request library for interacting with API
from getpass import getpass # Allows password entry
from .document import Document
from .utils import getSelection, prompt, color_by_status, parse, make_acronym
import readline
from .completer import Completer, readline_init
from builtins import input # So that Python 2 and 3 can both call input()

class PrintSession:
  # A printSession has a completer
  completer = Completer()

  # A printSession has a requestsSession for communicating with API
  session = requests.Session()

  # A print session has a buildings menu 
  buildings = None
  # A print session has a floors menu (populated after building selection)
  floors = None
  # A print session has a queues menu (populated after floor selection)
  queues = None

  # Building obj for user's desired building
  building = None
  # Floor obj for user's desired floor; within building 
  floor = None
  # Queue obj for user's desired floor
  queue = None

  # Print session has list of Document objects
  documents = []

  # Dictionary; Print session has command line arguments:
  # building 
  # noscale
  # documents ------- list
  # copies ---------- list
  # floorId
  # one
  # range ------------ list
  # short
  # landscape
  sysArgs = None

  def __init__(self):
    self.populateMenus()
    self.setSysArgs(vars(parse()))
    self.interpretSysArgs()
    self.authenticate()

  def setSysArgs(self, sysArgsIn):
    self.sysArgs = sysArgsIn

  def interpretSysArgs(self):
    # Sets Document params based on sys args
    if self.sysArgs['landscape']:
      Document.params['orientation'] = Document.LANDSCAPE
    if self.sysArgs['one']:
      Document.params['sides'] = Document.ONE_SIDED
    if self.sysArgs['short']:
      Document.params['sided'] = Document.TWO_SIDED_SHORT
    if self.sysArgs['noscale']:
      Document.params['scale'] = Document.OFF;
    if self.sysArgs['range']: 
      if len(self.sysArgs['range']) != len(self.sysArgs['documents']):
        sys.exit('If specifying ranges, must do so for all documents')
    if self.sysArgs['copies']: 
      if len(self.sysArgs['copies']) != len(self.sysArgs['documents']):
        sys.exit('If specifying copies, must do so for all documents')

    # Checks that documents exist and appends them to documents with ranges and copies
    if self.sysArgs['documents']:
      for i, pathIn in enumerate(self.sysArgs['documents']):
        if not os.path.isfile(pathIn):
          sys.exit('\'' + pathIn + '\' is not a file.')
        if not pathIn.lower().endswith(Document.VALID_EXTENSIONS):
          sys.exit('Invalid file extension')
        doc = Document(pathIn)
        if self.sysArgs['copies']:
          if not self.sysArgs['copies'][i].isdigit():
            sys.exit('Copy arguments must be integers')
          doc.params['copies'] = int(self.sysArgs['copies'][i])
        if self.sysArgs['range']:
          start, end = self.sysArgs['range'][i].split('-')
          if not start.isdigit() or not end.isdigit() or int(end) < int(start):
            sys.exit('Range arguments must be of form \'a-b c-d...\'')
          doc.params['range'] = self.sysArgs['range'][i]
        self.documents.append(doc)

    if self.sysArgs['floorId']:
      self.floor = list(filter(lambda dict: dict['id'] == self.sysArgs['floorId'], self.floors['result']))[0]
      if not self.floor:
        sys.exit('Could not find specified floor ID.')
      buildingId = self.floor['building_id']
      self.building = list(filter(lambda dict: dict['id'] == buildingId, self.buildings['result']))[0]

  def authenticate(self):
    try:
      response = self.session.get(webloginBaseURL)
      Completer.set_no_complete()
      response = self.session.post(webloginBaseURL, 
                                   data={'login':input(prompt('Enter username: ')), 
                                   'password':getpass(prompt=prompt('Enter password: '))}, 
                                   allow_redirects=False)

      # Redirect code means successful login
      if response.status_code != 302:
        sys.exit('Invalid credentials.')
    except KeyboardInterrupt:
      sys.exit()
    except requests.exceptions.RequestException:
      sys.exit('Problem calling mprint API. Check internet connection.')
   
  def populateMenus(self):
    try:
      # Populates buildings menu
      responseBuildings = self.session.get(guestBaseURL + 'buildings')
      if responseBuildings.status_code != 200:
        sys.exit('Could not retrieve buildings')
      self.buildings = responseBuildings.json()

      # Populates floors menu with all floors TODO this is unneccessary?
      responseFloors = self.session.get(guestBaseURL + 'floors')
      if responseFloors.status_code != 200:
        sys.exit('Could not retrieve floors')
      self.floors = responseFloors.json()
    except KeyboardInterrupt:
      sys.exit()
    # except:
    except requests.exceptions.RequestException:
      sys.exit('Problem calling mprint API. Check internet connection.') 

  def determineBuilding(self):
    possibleBuildings = []
    self.completer.set_vocab_list([x['name'] for x in self.buildings['result']])
    inputString = input(prompt('Enter building name: '))
    for dictionary in self.buildings['result']:
      if inputString.lower() in dictionary['name'].lower() or inputString.lower() in dictionary['id'].lower() or inputString.lower() in make_acronym(dictionary['name']).lower():
        possibleBuildings.append(dictionary)
    if len(possibleBuildings) == 1:
      self.building = possibleBuildings[0]
      return
    self.building = getSelection('Select building', possibleBuildings, 'name', 'id')

  def determineFloor(self):
    try:
      possibleFloors = self.session.get(userBaseURL + 'floors', 
                                        params={'buildingId':self.building['id']}).json()['result']
      self.completer.deactivate()
      self.floor = getSelection('Select floor', possibleFloors, 'name', 'id')
      # If none selected, reset building choice too
      if not self.floor:
        self.building = None
    except KeyboardInterrupt:
      sys.exit()
    # except:
    except requests.exceptions.RequestException:
      sys.exit('Problem calling mprint API. Check internet connection.') 

  def determineQueue(self):
    try:
      possibleQueues = self.session.get(userBaseURL + 'queues', 
                                        params={'floor':self.floor['id']}).json()['result']
      self.completer.deactivate()
      self.queue = getSelection('Select printer', possibleQueues, 'display_name', 'model_name', 'color', 'tabloid')
      # If none selected, reset floor choice too
      if not self.queue:
        self.floor = None
    except KeyboardInterrupt:
      sys.exit()
    # except:
    except requests.exceptions.RequestException:
      sys.exit('Problem calling mprint API. Check internet connection.') 

  def determineDocs(self):
    if not self.documents:
      # readline.set_completer(Completer().completer)
      # readline.parse_and_bind('tab: complete')
      self.completer.set_path_completion()
      doc_strings = input(prompt('Enter document paths separated by spaces: ')).split()
      for doc_string in doc_strings:
        if not os.path.isfile(doc_string):
          sys.exit('\'' + doc_string + '\' is not a file')
        self.documents.append(Document(doc_string))

  def printDocs(self):
    try:
      self.summarizeJob() # Blocks until user input
      for doc in self.documents:
        doc_file = {'file': open(doc.filePath, 'rb')}
        doc_params = doc.params 
        doc_params['queue'] = self.queue['name']
        printStatus = self.session.post(userBaseURL + 'jobs', files=doc_file, params=doc_params).json()
        print(os.path.basename(doc.filePath) + ": " + color_by_status(printStatus['status']))
    except KeyboardInterrupt:
      sys.exit()
    # except:
    except requests.exceptions.RequestException:
      sys.exit('Problem calling mprint API. Check internet connection.') 

  def summarizeJob(self):
    print("Building: " + self.building['name'])
    print("Floor: " + self.floor['name'])
    print("Queue: " + self.queue['display_name'])
    print("Documents: ") 
    for doc in self.documents:
      print(doc) 
    self.completer.deactivate()
    if not self.sysArgs['quick']:
      response = input(prompt("Press enter to continue; anything to quit")) # TODO
      if response != "":
        sys.exit()

  def determineDestination(self):
    while not self.building or not self.floor or not self.queue:
      if not self.building:
        self.determineBuilding()
      if self.building and not self.floor:
        self.determineFloor()
      if self.building and self.floor and not self.queue:
        self.determineQueue()



#! /usr/bin/env python

import sys
from .printsession import PrintSession
from .document import Document
from colorama import init as colorama_init

# TODO
# Test floor code flags -- add more shortcut flags -- try to minimize the minimum input
  # flag for color printing, tabloid flag
  # building specifier flag, and floor number? 
# Add way to print code cheat sheet?

def main(): 
  try:
    # Lets colored output work on Windows
    colorama_init()
    # Create new session; authenticates, populates menu
    session = PrintSession()
    # Determine printer
    session.determineDestination()
    # Determine documents
    session.determineDocs()
    # Print documenrs
    session.printDocs()
  except KeyboardInterrupt:
    sys.exit()


# if __name__ == "__main__":
  # main()



from __future__ import print_function # So that python 2.6+ can use the end= syntax
import argparse
import sys
from builtins import input
from termcolor import colored

# Takes in a list of dictionaries, a prompt, and any number of keys,
# prints nice columns from the lists along with selection indexes, 
# prompts user for a selection and then returns that dictionary
# keysToPrint is a tuple
def getSelection(prompt_string, list, *keysToPrint):
  if not list:
    return None

  # Okay, we need to get len(max(values of all dicts ar key)) for ea. key
  keyMaxLens = {}
  for key in keysToPrint:
    keyMaxLens[key] = len(str(key))
    for dict in list:
      if len(str(dict[key])) > keyMaxLens[key]:
        keyMaxLens[key] = len(str(dict[key]))

  INDEX_WIDTH = 5
  print(colored("".join('INDEX'.ljust(INDEX_WIDTH)), attrs=['underline']), end=" ")
  for key in keysToPrint:
    print(colored("".join(key.upper().ljust(keyMaxLens[key])), attrs=['underline']), end=" ")
  print('')
  for i, dict in enumerate(list):
    print("".join(('[' + str(i) + ']').ljust(INDEX_WIDTH)), end=" ")
    for key in keysToPrint:
      print("".join(str(change_01_to_NY(dict[key]).ljust(keyMaxLens[key]))), end=" ")
    print('')

  selection = input(prompt(prompt_string + ': '))
  try:
    selection_index = int(selection)
    if selection_index < 0: return None
    dict = list[selection_index]
    return dict
  except:
    return None

def change_01_to_NY(string):
  if string == '0':
    return 'N'
  elif string == '1':
    return 'Y'
  else:
    return string

def color_by_status(string):
  if 'success' in string: 
    return colored(string, 'green')
  else: 
    return colored(string, 'red')

# Escape sequences used for colors, in combination with readline, break input()
# Since tab completion is more useful than the bold prompt, the bold prompt is nixed
def prompt(string):
  # return colored(string, attrs=['bold'])
  return string

def parse():
  parser = argparse.ArgumentParser()
  parser.add_argument('-d', 
                      '--documents', 
                      nargs='+', 
                      required=False,
                      help='Specify at least one document to be printed')
  parser.add_argument('-f', 
                      '--floorId', 
                      action='store', 
                      required=False, 
                      help='Specify floorID of desired printer, if known')
  parser.add_argument('-b', 
                      '--building',
                      action='store',
                      required=False,
                      help='Specify building of desired printer')
  parser.add_argument('-l', 
                      '--landscape', 
                      action='store_const', 
                      const=True, 
                      required=False, 
                      help='Print document with landscape orientation (portrait is default)')
  parser.add_argument('-r',
                      '--range', 
                      nargs='+',
                      action='store',
                      required=False,
                      help='Print only pages within range; specify as \'x-x y-y z-z...\', one pair for each document')
  parser.add_argument('-c',
                      '--copies', 
                      nargs='+', 
                      required=False,
                      help='Specify copies to be printed of each document in corresponding order')
  parser.add_argument('-o', 
                      '--one',
                      action='store_const', 
                      const=True, 
                      required=False,
                      help='One-sided printing')
  parser.add_argument('-s', 
                      '--short',
                      action='store_const', 
                      const=True, 
                      required=False,
                      help='Double-sided short-edge printing')
  parser.add_argument('-n',
                      '--noscale',
                      action='store_const',
                      const=True,
                      required=False, 
                      help='Do not scale to page (default is fit to page)')
  parser.add_argument('-q',
                      '--quick',
                      action='store_const',
                      const=True,
                      required=False,
                      help='Quick mode: will not ask for confirmation before printing')
  args = parser.parse_args()
  return args

def make_acronym(string):
  return "".join(word[0] for word in string.replace('-', ' ').split() if word[0].isupper())







import readline
import sys
import os

class Completer:
  vocab_list = None

  # def __init__(self, vocab_list_in = None):
    # self.vocab_list = vocab_list_in

  def set_vocab_list(self, vocab_list_in = None):
    self.vocab_list = vocab_list_in
    readline.parse_and_bind('tab: complete')
    # Space must not be a delimiter since building names etc. can have spaces in them, and one is always specified
    readline.set_completer_delims('\t\n`!@#$%^&*()-=+[{]}\\|;:\'",<>?')
    readline.set_completer(self.completer)

  def set_path_completion(self):
    self.vocab_list = None
    readline.parse_and_bind('tab: complete')
    # Space must be a delimiter so that multiple docs can be specified
    readline.set_completer_delims(' \t\n`!@#$%^&*()-=+[{]}\\|;:\'",<>?') 
    readline.set_completer(self.completer)
    
  def deactivate(self):
    vocab_list = None
    readline.parse_and_bind('tab: self-insert')

  def list_from_path(self, path):
    if path.startswith(os.path.sep) or path.startswith('~'): # Absolute path 
      basedir = os.path.dirname(os.path.expanduser(path))
      contents = os.listdir(basedir)
      contents = [os.path.join(basedir, d) for d in contents]
    else: # Relative path
      if os.path.dirname(path) == '': # No directory yet specified
        contents = os.listdir(os.curdir)
      else: # At least one directory specified
        contents = os.listdir(os.path.dirname(path))
        contents = [os.path.join(os.path.dirname(path), d) for d in contents]
    contents = [d + os.path.sep if os.path.isdir(d) else d for d in contents]
    return contents

  def completer(self, text, state):
    if self.vocab_list:
      options = [x for x in self.vocab_list if x.lower().startswith(text.lower())]
      return options[state]
    options = [x for x in self.list_from_path(text) if x.startswith(os.path.expanduser(text))]
    return options[state]

  @staticmethod
  def set_no_complete():
    readline.set_completer(None)
    readline.parse_and_bind('tab: self-insert')
    readline.set_completer_delims(' \t\n`!@#$%^&*()-=+[{]}\\|;:\'",<>?')

def readline_init():
  print('running readline init')
  readline.parse_and_bind('tab: complete')
  readline.set_completer_delims(' \t\n`!@#$%^&*()-=+[{]}\\|;:\'",<>?')



#! /usr/bin/env python
from utils import prompt
from colorama import init as colorama_init
from builtins import input

def main():
  a = input(prompt('Enter something: '))

if __name__ == "__main__":
  main()



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='mprint-assistant', 
      version='0.1',
      description='Easy printing to U of M printers',
      long_description=long_description,
      author='Sam Adams',
      author_email='sam.h.adams@gmail.com',
      url='https://github.com/samharad/mprint-assistant',
      license='MIT',
      packages=['mprintassistant'],
      entry_points = {
        'console_scripts': ['mprint = mprintassistant.mprintassistant:main']
        },
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Printing'
        ],
      keywords='Michigan print mprint',
      install_requires=['colorama',
                        'termcolor',
                        'requests',
                        'readline',
                        'argparse',
                        'future'
                        ] 
      )
